"""Prepare a GeoDataFrame of line geometries for directed network analysis."""

import warnings

import pandas as pd
from geopandas import GeoDataFrame
from shapely.constructive import reverse

from ..helpers import return_two_vals, unit_is_meters


def make_directed_network_norway(gdf: GeoDataFrame) -> GeoDataFrame:
    """Runs the method make_directed_network for Norwegian road data.

    The parameters in make_directed_network are set to the correct values for
    Norwegian road data as of 2023.

    The data can be downloaded here:
    https://kartkatalog.geonorge.no/metadata/nvdb-ruteplan-nettverksdatasett/8d0f9066-34f9-4423-be12-8e8523089313

    Examples
    --------
    2022 data for the municipalities of Oslo and Eidskog can be read directly like this:

    >>> import sgis as sg
    >>> roads = sg.read_parquet_url(
    ...     "https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet"
    ... )
    >>> roads[["oneway", "drivetime_fw", "drivetime_bw", "geometry"]]
            oneway  drivetime_fw  drivetime_bw                                           geometry
    119702       B      0.216611      0.216611  MULTILINESTRING Z ((258028.440 6674249.890 413...
    199710      FT      0.099323     -1.000000  MULTILINESTRING Z ((271778.700 6653238.900 138...
    199725      FT      0.173963     -1.000000  MULTILINESTRING Z ((271884.510 6653207.540 142...
    199726      FT      0.011827     -1.000000  MULTILINESTRING Z ((271884.510 6653207.540 142...
    199733      FT      0.009097     -1.000000  MULTILINESTRING Z ((271877.373 6653214.697 141...
    ...        ...           ...           ...                                                ...
    1944129     FT      0.010482     -1.000000  MULTILINESTRING Z ((268649.503 6651869.320 111...
    1944392      B      0.135561      0.135561  MULTILINESTRING Z ((259501.570 6648459.580 3.2...
    1944398      B      0.068239      0.068239  MULTILINESTRING Z ((258292.600 6648313.440 18....
    1944409      B      0.023629      0.023629  MULTILINESTRING Z ((258291.452 6648289.258 19....
    1944415      B      0.175876      0.175876  MULTILINESTRING Z ((260762.830 6650240.620 43....
    [93395 rows x 46 columns]

    And converted to a directed network like this:

    >>> roads_directed = sg.make_directed_network_norway(roads)
    >>> roads_directed[["minutes", "geometry"]]
             minutes                                           geometry
    0       0.216611  MULTILINESTRING Z ((258028.440 6674249.890 413...
    1       0.028421  MULTILINESTRING Z ((266382.600 6639604.600 -99...
    2       0.047592  MULTILINESTRING Z ((266349.200 6639372.499 171...
    3       0.026180  MULTILINESTRING Z ((266360.515 6639120.493 172...
    4       0.023978  MULTILINESTRING Z ((266351.263 6639416.154 169...
    ...          ...                                                ...
    175620  0.007564  MULTILINESTRING Z ((268641.225 6651871.624 111...
    175621  0.020246  MULTILINESTRING Z ((268669.748 6651890.291 110...
    175622  0.036810  MULTILINESTRING Z ((268681.757 6651886.457 110...
    175623  0.003019  MULTILINESTRING Z ((268682.748 6651886.162 110...
    175624  0.036975  MULTILINESTRING Z ((268694.594 6651881.688 111...
    [175541 rows x 45 columns]
    """
    gdf = gdf.loc[~((gdf.drivetime_fw == -1) & (gdf.drivetime_bw == -1))]
    return make_directed_network(
        gdf,
        direction_col="oneway",
        direction_vals_bft=("B", "FT", "TF"),
        minute_cols=("drivetime_fw", "drivetime_bw"),
    )


def make_directed_network(
    gdf: GeoDataFrame,
    direction_col: str,
    direction_vals_bft: tuple[str, str, str],
    minute_cols: tuple[str, str] | str | None = None,
    speed_col_kmh: str | None = None,
    flat_speed_kmh: int | None = None,
    reverse_tofrom: bool = True,
) -> GeoDataFrame:
    """Flips the line geometries of roads going backwards and in both directions.

    Args:
        gdf: GeoDataFrame of line geometries.
        direction_col: name of column specifying the direction of the line
            geometry.
        direction_vals_bft: tuple or list with the values of the direction
            column. Must be in the order 'both directions', 'from', 'to'.
            E.g. ('B', 'F', 'T').
        minute_cols (optional): column or columns containing the number of minutes
            it takes to traverse the line. If one column name is given, this will
            be used for both directions. If tuple/list with two column names,
            the first column will be used as the minute column for the forward
            direction, and the second column for roads going backwards.
        speed_col_kmh (optional): name of column with the road speed limit.
        flat_speed_kmh (optional): Speed in kilometers per hour to use as the speed for
            all roads.
        reverse_tofrom: If the geometries of the lines going backwards
            (i.e. has the last value in 'direction_vals_bft'). Defaults to True.

    Returns:
        The Network class, with the network attribute updated with flipped
            geometries for lines going backwards and both directions.
        Adds the column 'minutes' if either 'speed_col_kmh', 'minute_col' or
            'flat_speed_kmh' is specified.

    Raises:
        ValueError: If 'flat_speed_kmh' or 'speed_col_kmh' is specified and the unit of
            the coordinate reference system is not 'metre'.
    """
    _validate_minute_args(minute_cols, speed_col_kmh, flat_speed_kmh)
    _validate_direction_args(gdf, direction_col, direction_vals_bft)

    if (flat_speed_kmh or speed_col_kmh) and not unit_is_meters(gdf):
        raise ValueError(
            "The crs must have 'metre' as units when calculating minutes."
            "Change crs or calculate minutes manually."
        )

    b, f, t = direction_vals_bft

    if minute_cols and minute_cols != "minutes" and minute_cols[0] != "minutes":
        gdf = gdf.drop("minutes", axis=1, errors="ignore")

    # select the directional and bidirectional rows.
    ft = gdf.loc[gdf[direction_col] == f]
    tf = gdf.loc[gdf[direction_col] == t]
    both_ways = gdf.loc[gdf[direction_col] == b]
    both_ways2 = both_ways.copy()

    if minute_cols:
        try:
            min_f, min_t = return_two_vals(minute_cols)
        except ValueError as e:
            raise ValueError(
                "'minute_cols' should be column name (string) or tuple/list with "
                "values of directions forwards and backwards, in that order."
            ) from e

        # rename the two minute cols
        both_ways = both_ways.rename(columns={min_f: "minutes"}, errors="raise")
        both_ways2 = both_ways2.rename(columns={min_t: "minutes"}, errors="raise")

        ft = ft.rename(columns={min_f: "minutes"}, errors="raise")
        tf = tf.rename(columns={min_t: "minutes"}, errors="raise")

    both_ways2.geometry = reverse(both_ways2.geometry)

    if reverse_tofrom:
        tf.geometry = reverse(tf.geometry)

    gdf = pd.concat([both_ways, both_ways2, ft, tf], ignore_index=True)

    if minute_cols and minute_cols != "minutes" and minute_cols[0] != "minutes":
        gdf = gdf.drop([min_f, min_t], axis=1, errors="ignore")

    if speed_col_kmh:
        gdf = _get_speed_from_col(gdf, speed_col_kmh)

    if flat_speed_kmh:
        meters_per_min = (flat_speed_kmh / 60) * 1000
        gdf["minutes"] = gdf.length / meters_per_min

    return gdf


def _validate_minute_args(minute_cols, speed_col_kmh, flat_speed_kmh):
    if not minute_cols and not speed_col_kmh and not flat_speed_kmh:
        warnings.warn(
            "Minute column will not be calculated when both 'minute_cols', "
            "'speed_col_kmh' and 'flat_speed_kmh' is None",
            stacklevel=2,
        )

    if sum([bool(minute_cols), bool(speed_col_kmh), bool(flat_speed_kmh)]) > 1:
        raise ValueError(
            "Can only calculate minutes from either 'speed_col_kmh', "
            "'minute_cols' or 'flat_speed_kmh'."
        )


def _validate_direction_args(gdf, direction_col, direction_vals_bft):
    if len(direction_vals_bft) != 3:
        raise ValueError(
            "'direction_vals_bft' should be tuple/list with values of directions "
            "both, from and to. E.g. ('B', 'F', 'T')"
        )

    b, f, t = direction_vals_bft

    if not all(gdf.loc[gdf[direction_col].isin(direction_vals_bft)]):
        raise ValueError(
            f"direction_col '{direction_col}' should have only three unique",
            f"values. unique values in {direction_col}:",
            ", ".join(gdf[direction_col].fillna("nan").unique()),
            f"Got {direction_vals_bft}.",
        )

    if "b" in t.lower() and "t" in b.lower() and "f" in f.lower():
        warnings.warn(
            f"The 'direction_vals_bft' should be in the order 'both ways', "
            f"'from', 'to'. Got {b, f, t}. Is this correct?",
            stacklevel=2,
        )


def _get_speed_from_col(gdf: GeoDataFrame, speed_col_kmh: str) -> GeoDataFrame:
    if len(gdf.loc[(gdf[speed_col_kmh].isna()) | (gdf[speed_col_kmh] == 0)]):
        raise ValueError(
            f"speed_col_kmh {speed_col_kmh!r} cannot have missing values or zeros"
        )

    gdf["minutes"] = gdf.length / (gdf[speed_col_kmh].astype(float) * 1000 / 60)

    return gdf



import numbers
import warnings
from collections.abc import Iterator, Sized

import geopandas as gpd
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from geopandas.array import GeometryDtype
from pandas.api.types import is_dict_like
from shapely import (
    Geometry,
    force_2d,
    get_exterior_ring,
    get_interior_ring,
    get_num_interior_rings,
    get_parts,
    wkb,
    wkt,
)
from shapely.geometry import LineString, Point
from shapely.ops import unary_union


def coordinate_array(
    gdf: GeoDataFrame | GeoSeries,
) -> np.ndarray[np.ndarray[float], np.ndarray[float]]:
    """Creates a 2d ndarray of coordinates from point geometries.

    Args:
        gdf: GeoDataFrame or GeoSeries of point geometries.

    Returns:
        np.ndarray of np.ndarrays of coordinates.

    Examples
    --------
    >>> from sgis import coordinate_array, random_points
    >>> points = random_points(5)
    >>> points
                    geometry
    0  POINT (0.59376 0.92577)
    1  POINT (0.34075 0.91650)
    2  POINT (0.74841 0.10627)
    3  POINT (0.00966 0.87868)
    4  POINT (0.38046 0.87879)
    >>> coordinate_array(points)
    array([[0.59376221, 0.92577159],
        [0.34074678, 0.91650446],
        [0.74840912, 0.10626954],
        [0.00965935, 0.87867915],
        [0.38045827, 0.87878816]])
    >>> coordinate_array(points.geometry)
    array([[0.59376221, 0.92577159],
        [0.34074678, 0.91650446],
        [0.74840912, 0.10626954],
        [0.00965935, 0.87867915],
        [0.38045827, 0.87878816]])
    """
    if isinstance(gdf, GeoDataFrame):
        return np.array([(geom.x, geom.y) for geom in gdf.geometry])
    else:
        return np.array([(geom.x, geom.y) for geom in gdf])


def _push_geom_col(gdf: GeoDataFrame) -> GeoDataFrame:
    """Makes the geometry column the rightmost column in the GeoDataFrame.

    Args:
        gdf: GeoDataFrame.

    Returns:
        The GeoDataFrame with the geometry column pushed all the way to the right.
    """
    geom_col = gdf._geometry_column_name
    return gdf.reindex(columns=[c for c in gdf.columns if c != geom_col] + [geom_col])


def drop_inactive_geometry_columns(gdf: GeoDataFrame) -> GeoDataFrame:
    for col in gdf.columns:
        if (
            isinstance(gdf[col].dtype, GeometryDtype)
            and col != gdf._geometry_column_name
        ):
            gdf = gdf.drop(col, axis=1)
    return gdf


def rename_geometry_if(gdf: GeoDataFrame) -> GeoDataFrame:
    geom_col = gdf._geometry_column_name
    if geom_col == "geometry" and geom_col in gdf.columns:
        return gdf
    elif geom_col in gdf.columns:
        return gdf.rename_geometry("geometry")

    geom_cols = list(
        {col for col in gdf.columns if isinstance(gdf[col].dtype, GeometryDtype)}
    )
    if len(geom_cols) == 1:
        gdf._geometry_column_name = geom_cols[0]
        return gdf.rename_geometry("geometry")

    raise ValueError(
        "There are multiple geometry columns and none are the active geometry"
    )


def clean_geoms(
    gdf: GeoDataFrame | GeoSeries, ignore_index: bool = False
) -> GeoDataFrame | GeoSeries:
    """Fixes geometries and removes invalid, empty, NaN and None geometries.

    Args:
        gdf: GeoDataFrame or GeoSeries to be cleaned.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False

    Returns:
        GeoDataFrame or GeoSeries with fixed geometries and only the rows with valid,
        non-empty and not-NaN/-None geometries.

    Examples
    --------
    >>> from sgis import clean_geoms, to_gdf
    >>> import pandas as pd
    >>> from shapely import wkt
    >>> gdf = to_gdf([
    ...         "POINT (0 0)",
    ...         "LINESTRING (1 1, 2 2)",
    ...         "POLYGON ((3 3, 4 4, 3 4, 3 3))"
    ...         ])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    Removing None and empty geometries.

    >>> missing = pd.DataFrame({"geometry": [None]})
    >>> empty = to_gdf(wkt.loads("POINT (0 0)").buffer(0))
    >>> gdf = pd.concat([gdf, missing, empty])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....
    0                                               None
    0                                      POLYGON EMPTY
    >>> clean_geoms(gdf)
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....
    """
    warnings.filterwarnings("ignore", "GeoSeries.notna", UserWarning)

    if isinstance(gdf, GeoDataFrame):
        geom_col = gdf._geometry_column_name
        gdf[geom_col] = gdf.make_valid()
        gdf = gdf.loc[
            (gdf[geom_col].is_valid)
            & (~gdf[geom_col].is_empty)
            & (gdf[geom_col].notna())
        ]
    elif isinstance(gdf, GeoSeries):
        gdf = gdf.make_valid()
        gdf = gdf.loc[(gdf.is_valid) & (~gdf.is_empty) & (gdf.notna())]
    else:
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    if ignore_index:
        gdf = gdf.reset_index(drop=True)

    return gdf


def random_points(n: int, loc: float | int = 0.5) -> GeoDataFrame:
    """Creates a GeoDataFrame with n random points.

    Args:
        n: Number of points/rows to create.
        loc: Mean ('centre') of the distribution.

    Returns:
        A GeoDataFrame of points with n rows.

    Examples
    --------
    >>> from sgis import random_points
    >>> points = random_points(10_000)
    >>> points
                         geometry
    0     POINT (0.62044 0.22805)
    1     POINT (0.31885 0.38109)
    2     POINT (0.39632 0.61130)
    3     POINT (0.99401 0.35732)
    4     POINT (0.76403 0.73539)
    ...                       ...
    9995  POINT (0.90433 0.75080)
    9996  POINT (0.10959 0.59785)
    9997  POINT (0.00330 0.79168)
    9998  POINT (0.90926 0.96215)
    9999  POINT (0.01386 0.22935)
    [10000 rows x 1 columns]

    Values with a mean of 100.

    >>> points = random_points(10_000, loc=100)
    >>> points
                         geometry
    0      POINT (50.442 199.729)
    1       POINT (26.450 83.367)
    2     POINT (111.054 147.610)
    3      POINT (93.141 141.456)
    4       POINT (94.101 24.837)
    ...                       ...
    9995   POINT (174.344 91.772)
    9996    POINT (95.375 11.391)
    9997    POINT (45.694 60.843)
    9998   POINT (73.261 101.881)
    9999  POINT (134.503 168.155)
    [10000 rows x 1 columns]
    """
    if isinstance(n, (str, float)):
        n = int(n)

    x = np.random.rand(n) * float(loc) * 2
    y = np.random.rand(n) * float(loc) * 2

    return GeoDataFrame(
        (Point(x, y) for x, y in zip(x, y, strict=True)), columns=["geometry"]
    )


def random_points_in_polygons(
    gdf: GeoDataFrame | GeoSeries, n: int, ignore_index=False
) -> GeoDataFrame:
    """Creates n random points inside each polygon of a GeoDataFrame.

    Args:
        gdf: GeoDataFrame to use as mask for the points.
        n: Number of points to create per polygon in 'gdf'.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False, meaning the points will have the index of the polygon
            it is within.

    Returns:
        A GeoDataFrame of points with 'n' rows per row in 'gdf'. It uses the index
        values of 'gdf'.

    Examples
    --------
    First create and buffer 100 random points.

    >>> import sgis as sg
    >>> gdf = sg.random_points(100)
    >>> polygons = sg.buff(gdf, 1)
    >>> polygons
                                                 geometry
    0   POLYGON ((1.49436 0.36088, 1.49387 0.32947, 1....
    1   POLYGON ((1.38427 0.21069, 1.38378 0.17928, 1....
    2   POLYGON ((1.78894 0.94134, 1.78845 0.90992, 1....
    3   POLYGON ((1.47174 0.81259, 1.47125 0.78118, 1....
    4   POLYGON ((1.13941 0.20821, 1.13892 0.17680, 1....
    ..                                                ...
    95  POLYGON ((1.13462 0.18908, 1.13412 0.15767, 1....
    96  POLYGON ((1.96391 0.43191, 1.96342 0.40050, 1....
    97  POLYGON ((1.30569 0.46956, 1.30520 0.43815, 1....
    98  POLYGON ((1.18172 0.10944, 1.18122 0.07803, 1....
    99  POLYGON ((1.06156 0.99893, 1.06107 0.96752, 1....
    [100 rows x 1 columns]

    >>> points = sg.random_points_in_polygons(polygons, 3)
    >>> points
                        geometry
    0   POINT (0.74944 -0.41658)
    0    POINT (1.27490 0.54076)
    0    POINT (0.22523 0.49323)
    1   POINT (0.25302 -0.34825)
    1    POINT (0.21124 0.89223)
    ..                       ...
    98  POINT (-0.39865 0.87135)
    98   POINT (0.03573 0.50788)
    99  POINT (-0.79089 0.57835)
    99   POINT (0.39838 1.50881)
    99   POINT (0.98383 0.77298)
    [300 rows x 1 columns]
    """
    if not isinstance(gdf, GeoDataFrame):
        gdf = to_gdf(gdf)

    if not all(gdf.geom_type.isin(["Polygon", "MultiPolygon"])):
        raise ValueError("Geometry types must be polygon.")

    if gdf.index.is_unique:
        gdf["temp_idx____"] = gdf.index
    else:
        gdf["temp_idx____"] = range(len(gdf))

    all_points = pd.DataFrame()

    for _ in range(n):
        bounds = gdf.bounds
        temp_idx____ = gdf["temp_idx____"].values
        overlapping = pd.DataFrame()
        overlapping_indices = ()

        while len(bounds):
            xs = np.random.uniform(bounds.minx, bounds.maxx)
            ys = np.random.uniform(bounds.miny, bounds.maxy)

            points_df = pd.DataFrame({"x": xs, "y": ys}, index=temp_idx____)

            points = to_gdf(points_df, geometry=["x", "y"], crs=gdf.crs).drop(
                ["x", "y"], axis=1
            )

            overlapping = points.sjoin(gdf[["temp_idx____", "geometry"]], how="inner")

            overlapping = overlapping.loc[overlapping.index == overlapping.temp_idx____]

            all_points = pd.concat([all_points, overlapping], ignore_index=ignore_index)

            overlapping_indices = overlapping_indices + tuple(overlapping.index.values)

            gdf__ = gdf.loc[~gdf["temp_idx____"].isin(overlapping_indices)]
            temp_idx____ = gdf__["temp_idx____"].values
            bounds = gdf__.bounds

    all_points = all_points.sort_index().drop(["temp_idx____", "index_right"], axis=1)

    if gdf.index.is_unique:
        gdf = gdf.drop("temp_idx____", axis=1)
        return all_points

    original_index = {
        temp_idx: idx for temp_idx, idx in zip(gdf.temp_idx____, gdf.index)
    }

    all_points.index = all_points.index.map(original_index)
    all_points.index.name = None

    gdf = gdf.drop("temp_idx____", axis=1)

    return all_points


def points_in_bounds(gdf: GeoDataFrame | GeoSeries, n2: int):
    minx, miny, maxx, maxy = gdf.total_bounds
    xs = np.linspace(minx, maxx, num=n2)
    ys = np.linspace(miny, maxy, num=n2)
    x_coords, y_coords = np.meshgrid(xs, ys, indexing="ij")
    coords = np.concatenate((x_coords.reshape(-1, 1), y_coords.reshape(-1, 1)), axis=1)
    return to_gdf(coords, crs=gdf.crs)


def to_lines(*gdfs: GeoDataFrame, copy: bool = True) -> GeoDataFrame:
    """Makes lines out of one or more GeoDataFrames and splits them at intersections.

    The GeoDataFrames' geometries are converted to LineStrings, then unioned together
    and made to singlepart. The lines are split at the intersections. Mimics
    'feature to line' in ArcGIS.

    Args:
        *gdfs: one or more GeoDataFrames.
        copy: whether to take a copy of the incoming GeoDataFrames. Defaults to True.

    Returns:
        A GeoDataFrame with singlepart line geometries and columns of all input
            GeoDataFrames.

    Note:
        The index is preserved if only one GeoDataFrame is given, but otherwise
        ignored. This is because the union overlay used if multiple GeoDataFrames
        always ignores the index.

    Examples
    --------
    Convert single polygon to linestring.

    >>> import sgis as sg
    >>> from shapely.geometry import Polygon
    >>> poly1 = sg.to_gdf(Polygon([(0, 0), (0, 1), (1, 1), (1, 0)]))
    >>> poly1["poly1"] = 1
    >>> line = sg.to_lines(poly1)
    >>> line
                                                geometry  poly1
    0  LINESTRING (0.00000 0.00000, 0.00000 1.00000, ...      1

    Convert two overlapping polygons to linestrings.

    >>> poly2 = sg.to_gdf(Polygon([(0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5)]))
    >>> poly2["poly2"] = 1
    >>> lines = sg.to_lines(poly1, poly2)
    >>> lines
    poly1  poly2                                           geometry
    0    1.0    NaN  LINESTRING (0.00000 0.00000, 0.00000 1.00000, ...
    1    1.0    NaN  LINESTRING (0.50000 1.00000, 1.00000 1.00000, ...
    2    1.0    NaN  LINESTRING (1.00000 0.50000, 1.00000 0.00000, ...
    3    NaN    1.0      LINESTRING (0.50000 0.50000, 0.50000 1.00000)
    4    NaN    1.0  LINESTRING (0.50000 1.00000, 0.50000 1.50000, ...
    5    NaN    1.0      LINESTRING (1.00000 0.50000, 0.50000 0.50000)

    Plot before and after.

    >>> sg.qtm(poly1, poly2)
    >>> lines["l"] = lines.length
    >>> sg.qtm(lines, "l")
    """

    if any(any(gdf.geom_type.isin(["Point", "MultiPoint"])) for gdf in gdfs):
        raise ValueError("Cannot convert points to lines.")

    def _shapely_geometry_to_lines(geom):
        if geom.area == 0:
            return geom

        singlepart = get_parts(geom)
        lines = []
        for part in singlepart:
            exterior_ring = get_exterior_ring(part)
            lines.append(exterior_ring)

            n_interior_rings = get_num_interior_rings(part)
            if not (n_interior_rings):
                continue

            interior_rings = [
                LineString(get_interior_ring(part, n)) for n in range(n_interior_rings)
            ]

            lines = lines + interior_rings

        return unary_union(lines)

    lines = []
    for gdf in gdfs:
        if copy:
            gdf = gdf.copy()

        gdf[gdf._geometry_column_name] = gdf[gdf._geometry_column_name].map(
            _shapely_geometry_to_lines
        )

        lines.append(gdf)

    if len(lines) == 1:
        return lines[0]

    unioned = lines[0].overlay(lines[1], how="union", keep_geom_type=True)

    if len(lines) > 2:
        for line_gdf in lines[2:]:
            unioned = unioned.overlay(line_gdf, how="union", keep_geom_type=True)

    return unioned.explode(ignore_index=True)


def to_multipoint(
    gdf: GeoDataFrame | GeoSeries, copy: bool = True
) -> GeoDataFrame | GeoSeries:
    """Creates multipoint geometries from GeoDataFrame or GeoSeries.

    Takes a GeoDataFrame or GeoSeries and turns it into a MultiPoint.

    Args:
        gdf: The geometry to be converted to MultiPoint.
        copy: If True, the geometry will be copied. Defaults to True.

    Returns:
        A GeoDataFrame or GeoSeries with MultiPoint geometries. If the input type
        if GeoDataFrame, the other columns will be preserved.

    Examples
    --------
    Let's create a GeoDataFrame with a point, a line and a polygon.

    >>> from sgis import to_multipoint, to_gdf
    >>> from shapely.geometry import LineString, Polygon
    >>> gdf = to_gdf([
    ...     (0, 0),
    ...     LineString([(1, 1), (2, 2)]),
    ...     Polygon([(3, 3), (4, 4), (3, 4), (3, 3)])
    ...     ])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    >>> to_multipoint(gdf)
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      MULTIPOINT (1.00000 1.00000, 2.00000 2.00000)
    2  MULTIPOINT (3.00000 3.00000, 3.00000 4.00000, ...
    """
    if copy:
        gdf = gdf.copy()

    if gdf.is_empty.any():
        raise ValueError("Cannot create multipoints from empty geometry.")

    def _to_multipoint(gdf):
        koordinater = "".join(
            [x for x in gdf.wkt if x.isdigit() or x.isspace() or x == "." or x == ","]
        ).strip()

        alle_punkter = [
            wkt.loads(f"POINT ({punkt.strip()})") for punkt in koordinater.split(",")
        ]

        return unary_union(alle_punkter)

    if isinstance(gdf, GeoDataFrame):
        gdf[gdf._geometry_column_name] = (
            gdf[gdf._geometry_column_name]
            .pipe(force_2d)
            .apply(lambda x: _to_multipoint(x))
        )

    elif isinstance(gdf, gpd.GeoSeries):
        gdf = force_2d(gdf)
        gdf = gdf.apply(lambda x: _to_multipoint(x))

    else:
        gdf = to_gdf(gdf)
        gdf["geometry"] = (
            gdf["geometry"].pipe(force_2d).apply(lambda x: _to_multipoint(x))
        )

    return gdf


def clean_clip(
    gdf: GeoDataFrame | GeoSeries,
    mask: GeoDataFrame | GeoSeries | Geometry,
    **kwargs,
) -> GeoDataFrame | GeoSeries:
    """Clips geometries to the mask extent and cleans the geometries.

    Geopandas.clip does a fast and durty clipping, with no guarantee for valid outputs.
    Here, the clipped geometries are made valid, and then empty, NaN and invalid
    geometries are removed.

    Args:
        gdf: GeoDataFrame or GeoSeries to be clipped
        mask: the geometry to clip gdf
        **kwargs: Additional keyword arguments passed to GeoDataFrame.clip

    Returns:
        The cleanly clipped GeoDataFrame.

    Raises:
        TypeError: If gdf is not of type GeoDataFrame or GeoSeries.
    """
    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    try:
        return gdf.clip(mask, **kwargs).pipe(clean_geoms)
    except Exception:
        gdf = clean_geoms(gdf)
        mask = clean_geoms(mask)
        return gdf.clip(mask, **kwargs).pipe(clean_geoms)


def to_gdf(
    geom: Geometry
    | str
    | bytes
    | list
    | tuple
    | dict
    | GeoSeries
    | pd.Series
    | pd.DataFrame
    | Iterator,
    crs: str | tuple[str] | None = None,
    geometry: str | None = None,
    **kwargs,
) -> GeoDataFrame:
    """Converts geometry-like objects to a GeoDataFrame.

    Constructs a GeoDataFrame from any geometry-like object, or an interable of such.
    Accepted types are string (wkt), byte (wkb), coordinate tuples, shapely geometries,
    GeoSeries, Series/DataFrame. The index/keys will be preserved if the input type is
    Series, DataFrame or dictionary.

    Args:
        geom: the object to be converted to a GeoDataFrame
        crs: if None (default), it uses the crs of the GeoSeries if GeoSeries
            is the input type. Otherwise, no crs is used.
        geometry: name of column(s) containing the geometry-like values. Can be
            ['x', 'y', 'z'] if coordinate columns, or e.g. 'geometry' if one column.
            The resulting geometry column will always be named 'geometry'.
        **kwargs: additional keyword arguments taken by the GeoDataFrame constructor.

    Returns:
        A GeoDataFrame with one column, the geometry column.

    Raises:
        TypeError: If geom is a GeoDataFrame.

    Examples
    --------
    >>> from sgis import to_gdf
    >>> coords = (10, 60)
    >>> to_gdf(coords, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)

    From wkt.

    >>> wkt = "POINT (10 60)"
    >>> to_gdf(wkt, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)

    From DataFrame with x, y (optionally z) coordinate columns. Index and
    columns are preserved.

    >>> df = pd.DataFrame({"x": [10, 11], "y": [60, 59]}, index=[1,3])
        x   y
    1  10  60
    3  11  59
    >>> gdf = to_gdf(df, geometry=["x", "y"], crs=4326)
    >>> gdf
        x   y                   geometry
    1  10  60  POINT (10.00000 60.00000)
    3  11  59  POINT (11.00000 59.00000)

    For DataFrame/dict with a geometry-like column, the geometry column can be
    speficied with the geometry parameter, which is set to "geometry" by default.

    >>> df = pd.DataFrame({"col": [1, 2], "geometry": ["point (10 60)", (11, 59)]})
    >>> df
       col       geometry
    0    1  point (10 60)
    1    2       (11, 59)
    >>> gdf = to_gdf(df, crs=4326)
    >>> gdf
       col                   geometry
    0    1  POINT (10.00000 60.00000)
    1    2  POINT (11.00000 59.00000)

    From Series or Series-like dictionary.

    >>> d = {1: (10, 60), 3: (11, 59)}
    >>> to_gdf(d)
                        geometry
    1  POINT (10.00000 60.00000)
    3  POINT (11.00000 59.00000)

    >>> from pandas import Series
    >>> to_gdf(Series(d))
                        geometry
    1  POINT (10.00000 60.00000)
    3  POINT (11.00000 59.00000)

    Multiple coordinates will be converted to points, unless a line or polygon geometry
    is constructed beforehand.

    >>> coordslist = [(10, 60), (11, 59)]
    >>> to_gdf(coordslist, crs=4326)
                        geometry
    0  POINT (10.00000 60.00000)
    1  POINT (11.00000 59.00000)

    >>> from shapely.geometry import LineString
    >>> to_gdf(LineString(coordslist), crs=4326)
                                                geometry
    0  LINESTRING (10.00000 60.00000, 11.00000 59.00000)

    From 2 or 3 dimensional array.

    >>> arr = np.random.randint(100, size=(5, 3))
    >>> to_gdf(arr)
                             geometry
    0  POINT Z (82.000 88.000 82.000)
    1  POINT Z (70.000 92.000 20.000)
    2   POINT Z (91.000 34.000 3.000)
    3   POINT Z (1.000 50.000 77.000)
    4  POINT Z (58.000 49.000 46.000)
    """
    if isinstance(geom, GeoDataFrame):
        raise TypeError("'to_gdf' doesn't accept GeoDataFrames as input type.")

    if isinstance(geom, GeoSeries):
        if not crs:
            crs = geom.crs
        else:
            geom = geom.to_crs(crs) if geom.crs else geom.set_crs(crs)

        return GeoDataFrame({"geometry": geom}, geometry="geometry", crs=crs, **kwargs)

    # first the iterators that get consumed by 'all' statements
    if isinstance(geom, Iterator) and not isinstance(geom, Sized):
        geom = GeoSeries(_make_shapely_geom(x) for x in geom)
        return GeoDataFrame({"geometry": geom}, geometry="geometry", crs=crs, **kwargs)

    # dataframes and dicts with geometry/xyz key(s)
    geometry = "geometry" if not geometry else geometry
    if _is_df_like(geom, geometry):
        geom = geom.copy()
        geom = _make_geomcol_df_like(geom, geometry, index=kwargs.get("index"))
        if "geometry" in geom:
            return GeoDataFrame(geom, geometry="geometry", crs=crs, **kwargs)
        elif isinstance(geom, pd.DataFrame):
            raise ValueError(f"Cannot find 'geometry' column {geometry!r}")

    if is_dict_like(geom):
        if isinstance(geom, dict):
            geom = pd.Series(geom)
        if "index" not in kwargs:
            index_ = geom.index
        else:
            index_ = None
        geom = GeoSeries(_make_shapely_geom(x) for x in geom)
        gdf = GeoDataFrame({"geometry": geom}, geometry="geometry", crs=crs, **kwargs)
        if index_ is not None:
            gdf.index = index_
        return gdf

    # single geometry objects like wkt, wkb, shapely geometry or iterable of numbers
    if _is_one_geometry(geom):
        geom = GeoSeries(_make_shapely_geom(geom))
        return GeoDataFrame({"geometry": geom}, geometry="geometry", crs=crs, **kwargs)

        # single geometry objects like wkt, wkb, shapely geometry or iterable of numbers
    if hasattr(geom, "__iter__"):
        geom = GeoSeries(_make_shapely_geom(x) for x in geom)
        return GeoDataFrame({"geometry": geom}, geometry="geometry", crs=crs, **kwargs)

    raise TypeError(f"Got unexpected type {type(geom)}")


def _is_one_geometry(geom) -> bool:
    if (
        isinstance(geom, (str, bytes, Geometry))
        or all(isinstance(i, numbers.Number) for i in geom)
        or not hasattr(geom, "__iter__")
    ):
        return True
    return False


def _is_df_like(geom, geometry: str | None) -> bool:
    if not is_dict_like(geom):
        return False

    if len(geometry) == 1 and geometry[0] in geom:
        return True

    if len(geometry) == 2:
        x, y = geometry
        if x in geom and y in geom:
            return True

    elif len(geometry) == 3:
        x, y, z = geometry
        if x in geom and y in geom and z in geom:
            return True

    if geometry in geom:
        return True

    if len(geom.keys()) == 1:
        return True


def _make_geomcol_df_like(
    geom: pd.DataFrame | dict,
    geometry: str | tuple[str],
    index: list | tuple | None,
) -> pd.DataFrame | dict:
    """Create GeoSeries column in DataFrame or dictionary that has a DataFrame structure
    rather than a Series structure.

    Use the same logic for DataFrame and dict, as long as the speficied 'geometry'
    value or values (if x, y, z coordinate columns) are keys in the dictionary. If
    not, nothing happens here, and the dict goes on to be treated as a Series where
    keys are index and values the geometries.
    """

    if not index and hasattr(geom, "index"):
        index = geom.index
    elif index and hasattr(geom, "index"):
        geom.index = index

    if len(geometry) == 1 and geometry[0] in geom:
        geometry = geometry[0]

    if isinstance(geometry, (tuple, list)):
        if len(geometry) == 2:
            x, y = geometry
            z = None
        elif len(geometry) == 3:
            x, y, z = geometry
            z = geom[z]
        else:
            raise ValueError(
                "geometry should be one geometry-like column or 2-3 x, y, z columns."
            )
        geom["geometry"] = gpd.GeoSeries.from_xy(x=geom[x], y=geom[y], z=z)
    elif geometry in geom:
        if not hasattr(geom[geometry], "__iter__"):
            geom[geometry] = [geom[geometry]]
        geom["geometry"] = GeoSeries(
            map(_make_shapely_geom, geom[geometry]), index=index
        )

    elif isinstance(geom, pd.DataFrame) and len(geom.columns) == 1:
        geometry = geom.columns[0]
        geom["geometry"] = GeoSeries(
            map(_make_shapely_geom, geom[geometry]), index=index
        )

    elif isinstance(geom, dict) and len(geom.keys()) == 1:
        geometry = next(iter(geom))
        geom["geometry"] = GeoSeries(
            map(_make_shapely_geom, geom[geometry]), index=index
        )

    return geom


def _make_shapely_geom(geom):
    """Create shapely geometry from wkt, wkb or coordinate tuple.

    Works recursively if the object is a nested iterable.
    """
    if isinstance(geom, str):
        return wkt.loads(geom)

    if isinstance(geom, bytes):
        return wkb.loads(geom)

    if isinstance(geom, Geometry):
        return geom

    if not hasattr(geom, "__iter__"):
        raise ValueError(
            f"Couldn't create shapely geometry from {geom} of type {type(geom)}"
        )

    if isinstance(geom, GeoSeries):
        raise TypeError(
            "to_gdf doesn't accept iterable of GeoSeries. Instead use: "
            "pd.concat(to_gdf(geom) for geom in geoseries_iterable)"
        )

    if not any(isinstance(g, numbers.Number) for g in geom):
        # we're likely dealing with a nested iterable, so let's
        # recursively dig out way down to the coords/wkt/wkb
        return unary_union([_make_shapely_geom(g) for g in geom])
    elif len(geom) == 2 or len(geom) == 3:
        return Point(geom)
    else:
        raise ValueError(
            "If 'geom' is an iterable, each item should consist of "
            "wkt, wkb or 2/3 coordinates (x, y, z)."
        )



"""Interactive map of one or more GeoDataFrames with layers that can be toggles on/off.

This module holds the Explore class, which is the basis for the explore, samplemap and
clipmap functions from the 'maps' module.
"""
import warnings

import matplotlib
import matplotlib.colors as colors
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame, GeoSeries
from jenkspy import jenks_breaks
from mapclassify import classify
from shapely import Geometry

from ..geopandas_tools.general import drop_inactive_geometry_columns, rename_geometry_if
from ..helpers import get_name


# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
pd.options.mode.chained_assignment = None


# custom default colors for non-numeric data, because the geopandas default has very
# similar colors. The palette is like the "Set2" cmap from matplotlib, but with more
# colors. If more than 14 categories, the geopandas default cmap is used.
_CATEGORICAL_CMAP = {
    0: "#4576ff",
    1: "#ff455e",
    2: "#59d45f",
    3: "#b51d8b",
    4: "#ffa514",
    5: "#f2dc4e",
    6: "#ff8cc9",
    7: "#6bf2eb",
    8: "#916209",
    9: "#008d94",
    10: "#8a030a",
    11: "#9c65db",
    12: "#228000",
    13: "#80ff00",
}


class Map:
    """Base class that prepares one or more GeoDataFrames for mapping.

    The class has no public methods.
    """

    def __init__(
        self,
        *gdfs: GeoDataFrame,
        column: str | None = None,
        labels: tuple[str] | None = None,
        k: int = 5,
        bins: tuple[float] | None = None,
        nan_label: str = "Missing",
        nan_color="#c2c2c2",
        **kwargs,
    ):
        if not all(isinstance(gdf, GeoDataFrame) for gdf in gdfs):
            gdfs, column = self._separate_args(gdfs, column)

        self._column = column
        self.bins = bins
        self._k = k
        self.nan_label = nan_label
        self.nan_color = nan_color
        self._cmap = kwargs.pop("cmap", None)

        if not all(isinstance(gdf, GeoDataFrame) for gdf in gdfs):
            raise ValueError("gdfs must be GeoDataFrames.")

        if not any(len(gdf) for gdf in gdfs):
            raise ValueError("None of the GeoDataFrames have rows.")

        if "namedict" in kwargs:
            for i, gdf in enumerate(gdfs):
                gdf.name = kwargs["namedict"][i]
            kwargs.pop("namedict")

        # need to get the object names of the gdfs before copying. Only getting,
        # not setting, labels. So the original gdfs don't get the label column.
        self.labels = labels
        if not self.labels:
            self._get_labels(gdfs)

        self._gdfs: list[GeoDataFrame] = [gdf.copy() for gdf in gdfs]
        self.kwargs = kwargs

        if not self.labels:
            self._set_labels()

        if not self._column:
            for gdf, label in zip(self._gdfs, self.labels, strict=True):
                gdf["label"] = label
            self._column = "label"

        self._gdfs = self._to_common_crs_and_one_geom_col(self._gdfs)
        self._is_categorical = self._check_if_categorical()

        self._fillna_if_col_is_missing()

        self._gdf = pd.concat(self._gdfs, ignore_index=True)

        self._nan_idx = self._gdf[self._column].isna()

        if not self._is_categorical:
            self._unique_values = self._get_unique_floats()
            if self._k > len(self._unique_values):
                self._k = len(self._unique_values)
        else:
            self._unique_values = sorted(
                list(self._gdf.loc[~self._nan_idx, self._column].unique())
            )

    def _get_unique_floats(self) -> np.array:
        """Get unique floats by multiplying, then converting to integer.

        Find a multiplier that makes the max value greater than +- 1_000_000.
        Because floats don't always equal each other. This will make very
        similar values count as the same value in the color classification.
        """
        array = self._gdf.loc[~self._nan_idx, self._column]

        self._min = np.min(array)
        self._max = np.max(array)
        self._get_multiplier(array)

        unique = array.reset_index(drop=True).drop_duplicates()
        as_int = self._array_to_large_int(unique)
        no_duplicates = as_int.drop_duplicates()

        return np.sort(np.array(unique.loc[no_duplicates.index]))

    def _array_to_large_int(self, array: np.ndarray):
        """Multiply values in float array, then convert to integer."""

        unique_multiplied = array * self._multiplier

        return unique_multiplied.astype(np.int64)

    def _get_multiplier(self, array: np.ndarray):
        """Find the number of zeros needed to push the max value of the array above
        +-1_000_000.

        Adding this as an attribute to use later in _classify_from_bins.
        """
        multiplier = 10
        max_ = np.max(array * multiplier)

        if self._max > 0:
            while max_ < 1_000_000:
                multiplier *= 10
                max_ = np.max(array * multiplier)
        else:
            while max_ > -1_000_000:
                multiplier *= 10
                max_ = np.max(array * multiplier)

        self._multiplier: int = multiplier

    def _add_minmax_to_bins(self, bins: list[float | int]) -> list[float | int]:
        """If values are outside the bin range, add max and/or min values of array."""
        # make sure they are lists
        bins = [bin for bin in bins]

        if min(bins) > 0 and min(self._gdf[self._column]) < min(bins):
            bins = [min(self._gdf[self._column])] + bins

        if min(bins) < 0 and min(self._gdf[self._column]) < min(bins):
            bins = [min(self._gdf[self._column])] + bins

        if max(bins) > 0 and max(self._gdf[self._column]) > max(bins):
            bins = bins + [max(self._gdf[self._column])]

        if max(bins) < 0 and max(self._gdf[self._column]) < max(bins):
            bins = bins + [max(self._gdf[self._column])]

        return bins

    @staticmethod
    def _separate_args(
        args: tuple,
        column: str | None,
    ) -> tuple[tuple[GeoDataFrame], str]:
        """Separate GeoDataFrames from string (column argument)."""

        gdfs: tuple[GeoDataFrame] = ()
        for arg in args:
            if isinstance(arg, str):
                if column is None:
                    column = arg
                else:
                    raise ValueError(
                        "Can specify at most one string as a positional argument."
                    )
            elif isinstance(arg, (GeoDataFrame, GeoSeries, Geometry)):
                gdfs = gdfs + (arg,)

        return gdfs, column

    def _prepare_continous_map(self):
        """Create bins if not already done and adjust k if needed."""

        if not hasattr(self, "scheme"):
            self.scheme = self.kwargs.pop("scheme", "fisherjenks")

        if self.scheme is None:
            return

        if not self.bins:
            self.bins = self._create_bins(self._gdf, self._column)
            if len(self.bins) <= self._k and len(self.bins) != len(self._unique_values):
                warnings.warn(f"Could not create {self._k} classes.")
                self._k = len(self.bins)
        else:
            self.bins = self._add_minmax_to_bins(self.bins)
            if len(self._unique_values) > len(self.bins):
                self._k = len(self.bins) - 1

    def _get_labels(self, gdfs: tuple[GeoDataFrame]) -> None:
        """Putting the labels/names in a list before copying the gdfs."""
        self.labels: list[str] = []
        for i, gdf in enumerate(gdfs):
            if hasattr(gdf, "name"):
                name = gdf.name
            else:
                name = get_name(gdf)
                if not name:
                    name = str(i)
            self.labels.append(name)

    def _set_labels(self) -> None:
        """Setting the labels after copying the gdfs."""
        for i, gdf in enumerate(self._gdfs):
            gdf["label"] = self.labels[i]

    def _to_common_crs_and_one_geom_col(self, gdfs: list[GeoDataFrame]):
        """Need common crs and max one geometry column."""
        crs_list = list({gdf.crs for gdf in gdfs if gdf.crs is not None})
        if crs_list:
            self.crs = crs_list[0]
        new_gdfs = []
        for gdf in gdfs:
            gdf = drop_inactive_geometry_columns(gdf).pipe(rename_geometry_if)
            if crs_list:
                try:
                    gdf = gdf.to_crs(self.crs)
                except ValueError:
                    gdf = gdf.set_crs(self.crs)
            new_gdfs.append(gdf)
        return new_gdfs

    def _fillna_if_col_is_missing(self) -> None:
        for gdf in self._gdfs:
            if self._column in gdf.columns:
                continue
            gdf[self._column] = pd.NA

    def _check_if_categorical(self) -> bool:
        """Quite messy this..."""
        if not self._column:
            return True

        maybe_area = 1 if "area" in self._column else 0
        maybe_length = (
            1 if any(x in self._column for x in ["meter", "metre", "leng"]) else 0
        )

        all_nan = 0
        col_not_present = 0
        for gdf in self._gdfs:
            if self._column not in gdf:
                if maybe_area:
                    gdf["area"] = gdf.area
                    maybe_area += 1
                elif maybe_length:
                    gdf["length"] = gdf.length
                    maybe_length += 1
                else:
                    col_not_present += 1
            elif not pd.api.types.is_numeric_dtype(gdf[self._column]):
                if all(gdf[self._column].isna()):
                    all_nan += 1
                return True

        if maybe_area > 1:
            self._column = "area"
            return False
        if maybe_length > 1:
            self._column = "length"
            return False

        if all_nan == len(self._gdfs):
            raise ValueError(f"All values are NaN in column {self.column!r}.")

        if col_not_present == len(self._gdfs):
            raise ValueError(f"{self.column} not found.")

        return False

    def _get_categorical_colors(self) -> None:
        # custom categorical cmap
        if not self._cmap and len(self._unique_values) <= len(_CATEGORICAL_CMAP):
            self._categories_colors_dict = {
                category: _CATEGORICAL_CMAP[i]
                for i, category in enumerate(self._unique_values)
            }
        elif self._cmap:
            cmap = matplotlib.colormaps.get_cmap(self._cmap)

            self._categories_colors_dict = {
                category: colors.to_hex(cmap(int(i)))
                for i, category in enumerate(self._unique_values)
            }
        else:
            cmap = matplotlib.colormaps.get_cmap("tab20")

            self._categories_colors_dict = {
                category: colors.to_hex(cmap(int(i)))
                for i, category in enumerate(self._unique_values)
            }

        if any(self._nan_idx):
            self._gdf[self._column] = self._gdf[self._column].fillna(self.nan_label)
            self._categories_colors_dict[self.nan_label] = self.nan_color

        for gdf in self._gdfs:
            gdf["color"] = gdf[self._column].map(self._categories_colors_dict)

        self._gdf["color"] = self._gdf[self._column].map(self._categories_colors_dict)

    def _create_bins(self, gdf: GeoDataFrame, column: str) -> np.ndarray:
        """Make bin list of length k + 1, or length of unique values.

        The returned bins sometimes have two almost identical

        If 'scheme' is not specified, the jenks_breaks function is used, which is
        much faster than the one from Mapclassifier.
        """

        n_classes = (
            self._k if len(self._unique_values) > self._k else len(self._unique_values)
        )

        if self._k == len(self._unique_values) - 1:
            n_classes = self._k - 1

        if self.scheme == "fisherjenks":
            bins = jenks_breaks(gdf.loc[~self._nan_idx, column], n_classes=n_classes)
        else:
            binning = classify(
                np.asarray(gdf.loc[~self._nan_idx, column]),
                scheme=self.scheme,
                k=self._k,
            )
            bins = binning.bins
            bins = self._add_minmax_to_bins(bins)

        unique_bins = list({round(bin, 5) for bin in bins})
        unique_bins.sort()

        if self._k == len(self._unique_values) - 1:
            return np.array(unique_bins)

        if len(unique_bins) == len(self._unique_values):
            return np.array(unique_bins)

        return np.array(bins)

    def change_cmap(self, cmap: str, start: int = 0, stop: int = 256):
        """Change the color palette of the plot.

        Args:
            cmap: The colormap.
                https://matplotlib.org/stable/tutorials/colors/colormaps.html
            start: Start position for the color palette. Defaults to 0.
            stop: End position for the color palette. Defaults to 256, which
                is the end of the color range.
        """
        self.cmap_start = start
        self.cmap_stop = stop
        self._cmap = cmap
        self._cmap_has_been_set = True
        return self

    def _get_continous_colors(self, n: int) -> np.ndarray:
        cmap = matplotlib.colormaps.get_cmap(self._cmap)
        colors_ = [
            colors.to_hex(cmap(int(i)))
            #            for i in np.linspace(self.cmap_start, self.cmap_stop, num=self._k)
            for i in np.linspace(self.cmap_start, self.cmap_stop, num=n)
        ]
        if any(self._nan_idx):
            colors_ = colors_ + [self.nan_color]
        return np.array(colors_)

    def _classify_from_bins(self, gdf: GeoDataFrame, bins: np.ndarray) -> np.ndarray:
        """Place the column values into groups."""
        if len(bins) == len(self._unique_values):
            # if equal lenght, convert to integer and check for equality
            gdf["col_as_int"] = self._array_to_large_int(gdf[self._column])
            bins = self._array_to_large_int(self._unique_values)
            classified = np.searchsorted(bins, gdf["col_as_int"])
        else:
            if len(bins) == self._k + 1:
                bins = bins[1:]

            classified = np.searchsorted(bins, gdf[self._column])

        return classified

    def _push_classification(self, classified: np.ndarray) -> np.ndarray:
        """Push classes downwards if gaps in classification sequence.

        So from e.g. [0,2,4] to [0,1,2].

        Otherwise, will get index error when classifying colors.
        """
        rank_dict = {val: rank for rank, val in enumerate(np.unique(classified))}

        return np.array([rank_dict[val] for val in classified])

    @property
    def k(self):
        return self._k

    @k.setter
    def k(self, new_value: bool):
        if not self._is_categorical and new_value > len(self._unique_values):
            raise ValueError(
                "'k' cannot be greater than the number of unique values in the column.'"
                f"Got new k={new_value} and previous k={len(self._unique_values)}.'"
                #  f"{''.join(self._unique_values)}"
            )
        self._k = int(new_value)

    @property
    def cmap(self):
        return self._cmap

    @cmap.setter
    def cmap(self, new_value: bool):
        self._cmap = new_value
        self.change_cmap(cmap=new_value, start=self.cmap_start, stop=self.cmap_stop)

    @property
    def gdf(self):
        return self._gdf

    @gdf.setter
    def gdf(self, _):
        raise ValueError(
            "Cannot change 'gdf' after init. Put the GeoDataFrames into "
            "the class initialiser."
        )

    @property
    def gdfs(self):
        return self._gdfs

    @gdfs.setter
    def gdfs(self, _):
        raise ValueError(
            "Cannot change 'gdfs' after init. Put the GeoDataFrames into "
            "the class initialiser."
        )

    @property
    def column(self):
        return self._column

    @column.setter
    def column(self, _):
        raise ValueError(
            "Cannot change 'column' after init. Specify 'column' in the "
            "class initialiser."
        )

    def __setitem__(self, item, new_item):
        return setattr(self, item, new_item)

    def __getitem__(self, item):
        return getattr(self, item)

    def get(self, key, default=None):
        try:
            return self[key]
        except (KeyError, ValueError, IndexError, AttributeError):
            return default



"""Functions for reading and writing GeoDataFrames in Statistics Norway's GCS Dapla.
"""
import os

import geopandas as gpd
import pandas as pd
from dapla import FileClient, details
from geopandas import GeoDataFrame
from geopandas.io.arrow import _geopandas_to_arrow
from pyarrow import parquet


def exists(path: str) -> bool:
    """Returns True if the path exists, and False if it doesn't.

    Works in dapla and outside of dapla.

    Args:
        path (str): The path to the file or directory.

    Returns:
        True if the path exists, False if not.
    """
    try:
        details(path)
        return True
    except FileNotFoundError:
        return False
    except ModuleNotFoundError:
        return os.path.exists(path)


def read_geopandas(path: str, **kwargs) -> GeoDataFrame:
    """Reads geoparquet or other geodata from a file on GCS.

    Note:
        Does not currently read shapefiles or filegeodatabases.

    Args:
        path: path to a file on Google Cloud Storage.
        **kwargs: Additional keyword arguments passed to geopandas' read_parquet
            or read_file, depending on the file type.

     Returns:
         A GeoDataFrame.
    """
    fs = FileClient.get_gcs_file_system()

    if "parquet" in path:
        with fs.open(path, mode="rb") as file:
            return gpd.read_parquet(file, **kwargs)
    else:
        with fs.open(path, mode="rb") as file:
            return gpd.read_file(file, **kwargs)


def write_geopandas(df: gpd.GeoDataFrame, gcs_path: str, **kwargs) -> None:
    """Writes a GeoDataFrame to the speficied format.

    Note:
        Does not currently write to shapelfile or filegeodatabase.

    Args:
        df: The GeoDataFrame to write.
        gcs_path: The path to the file you want to write to.
        **kwargs: Additional keyword arguments passed to parquet.write_table
            (for parquet) or geopandas' to_file method (if not parquet).
    """
    pd.io.parquet.BaseImpl.validate_dataframe(df)

    fs = FileClient.get_gcs_file_system()

    if ".parquet" in gcs_path:
        with fs.open(gcs_path, mode="wb") as buffer:
            table = _geopandas_to_arrow(df, index=df.index, schema_version=None)
            parquet.write_table(table, buffer, compression="snappy", **kwargs)
        return

    if ".gpkg" in gcs_path:
        driver = "GPKG"
    elif ".geojson" in gcs_path:
        driver = "GeoJSON"
    elif ".gml" in gcs_path:
        driver = "GML"
    elif ".shp" in gcs_path:
        driver = "ESRI Shapefile"
    else:
        driver = None

    with fs.open(gcs_path, "wb") as file:
        df.to_file(file, driver=driver)



# flake8: noqa: F401
from .geopandas_tools.buffer_dissolve_explode import (
    buff,
    buffdiss,
    buffdissexp,
    buffexp,
    dissexp,
)
from .geopandas_tools.general import (
    clean_clip,
    clean_geoms,
    coordinate_array,
    drop_inactive_geometry_columns,
    points_in_bounds,
    random_points,
    random_points_in_polygons,
    rename_geometry_if,
    to_gdf,
    to_lines,
    to_multipoint,
)
from .geopandas_tools.geometry_types import (
    get_geom_type,
    is_single_geom_type,
    to_single_geom_type,
)
from .geopandas_tools.neighbors import (
    get_all_distances,
    get_k_nearest_neighbors,
    get_neighbor_indices,
    k_nearest_neighbors,
)
from .geopandas_tools.overlay import clean_shapely_overlay, overlay, overlay_update
from .geopandas_tools.point_operations import snap_all, snap_within_distance
from .geopandas_tools.polygon_operations import close_all_holes, close_small_holes
from .maps.legend import Legend
from .maps.maps import clipmap, explore, qtm, samplemap
from .maps.thematicmap import ThematicMap
from .networkanalysis.closing_network_holes import (
    close_network_holes,
    close_network_holes_to_deadends,
)
from .networkanalysis.cutting_lines import (
    cut_lines,
    cut_lines_once,
    split_lines_by_nearest_point,
)
from .networkanalysis.directednetwork import (
    make_directed_network,
    make_directed_network_norway,
)
from .networkanalysis.finding_isolated_networks import (
    get_component_size,
    get_connected_components,
)
from .networkanalysis.network import Network
from .networkanalysis.networkanalysis import NetworkAnalysis
from .networkanalysis.networkanalysisrules import NetworkAnalysisRules
from .networkanalysis.nodes import (
    make_edge_coords_cols,
    make_edge_wkt_cols,
    make_node_ids,
)
from .read_parquet import read_parquet_url


try:
    from .dapla import exists, read_geopandas, write_geopandas
except ImportError:
    pass



"""Some small exception classes."""


class NoPointsWithinSearchToleranceError(Exception):
    """Exception for when the points are too far away from the network."""

    def __init__(
        self, what: str | None = None, search_tolerance: str | None = None
    ) -> None:
        self.what = what
        self.search_tolerance = search_tolerance

    def __str__(self):
        return (
            f"No {self.what} within specified 'search_tolerance' "
            f"of {self.search_tolerance}"
        )


class ZeroLinesError(Exception):
    """DataFrame has 0 rows."""


class NotInJupyterError(Exception):
    """This functionality only works in Jupyter."""



"""Small helper functions."""
import inspect
import warnings

from geopandas import GeoDataFrame


def return_two_vals(
    vals: tuple[str | None, str | None] | list[str] | str | int | float
) -> tuple[str | int | float, str | int | float | None]:
    """Return a two-length tuple from a str/int/float or list/tuple of length 1 or 2.

    Returns 'vals' as a 2-length tuple. If the input is a string, return
    a tuple of two strings. If the input is a list or tuple of two
    strings, return the list or tuple. Otherwise, raise a ValueError

    Args:
        vals: the item to be converted to a tuple of two values.

    Returns:
        A tuple of two strings, integers or floats.
    """
    if isinstance(vals, str):
        return vals, vals
    if hasattr(vals, "__iter__"):
        if len(vals) == 2:
            return vals[0], vals[1]
        if len(vals) == 1:
            return vals[0], vals[0]
        raise ValueError("list/tuple should be of length 1 or 2.")

    return vals, vals


def unit_is_meters(gdf: GeoDataFrame) -> bool:
    """Returns True if the crs unit is 'metre''."""
    unit = gdf.crs.axis_info[0].unit_name
    if unit != "metre":
        return False
    return True


def unit_is_metres(gdf: GeoDataFrame) -> bool:
    """Returns True if the crs unit is 'metre''."""
    return unit_is_meters(gdf)


def unit_is_degrees(gdf: GeoDataFrame) -> bool:
    """Returns True if the crs unit is 'metre''."""
    unit = gdf.crs.axis_info[0].unit_name
    if "degree" in unit:
        return True
    return False


def get_name(var: object, n: int = 5) -> str | None:
    """Searches through the local variables down one level at a time."""
    frame = inspect.currentframe().f_back.f_back

    for _ in range(n):
        locals_ = frame.f_locals

        names = [var_name for var_name, var_val in locals_.items() if var_val is var]
        if names and len(names) == 1:
            return names[0]

        names = [name for name in names if not name.startswith("_")]

        if names and len(names) == 1:
            return names[0]

        if names and len(names) > 1:
            warnings.warn(
                "More than one local variable matches the object. Name might be wrong."
            )
            return names[0]

        frame = frame.f_back

        if not frame:
            return


def make_namedict(gdfs: tuple[GeoDataFrame]) -> dict[int, str]:
    namedict = {}
    for i, gdf in enumerate(gdfs):
        if hasattr(gdf, "name"):
            namedict[i] = gdf.name
        else:
            name = get_name(gdf)
            if not name:
                name = str(i)
            namedict[i] = name
    return namedict
